//	Notes:
//		Not much to say here - might be nice to have a further abstraction for
//		handling containers of streams, for zip archives etc.  Zip archives do
//		require a check on filesize to pick out something over 100K that could
//		be a disk image, tho that doesn't really belong in a generic class
//
//		Zip disks are read only as ZLib only provides read-only access.  It'd
//		be relatively easy to add support for another 3rd party module to give
//		write access, tho it's probably not too important at the moment.

#include "stdafx.h"

#include "Util.h"
#include "CStream.h"
#include "Floppy.h"

////////////////////////////////////////////////////////////////////////////////

CStream::CStream (const char* pcszStream_, bool fReadOnly_/*=false*/)
{
	// No file open initially
	m_nMode = modeClosed;

	// Save whether we're to treat the stream as read-only
	m_fReadOnly = fReadOnly_;

	// Keep a copy of the stream source as we'll need it for saving
	m_pcszStream = strdup(pcszStream_);
}

/*virtual*/ CStream::~CStream ()
{
	free(const_cast<char*>(m_pcszStream));
}


// Identify the stream and create an object to supply data from it
/*static*/ CStream* CStream::Open (const char* pcszStream_, bool fReadOnly_/*=false*/)
{
	if (CFloppyStream::IsRecognised(pcszStream_))
		return new CFloppyStream(pcszStream_, fReadOnly_);
	else
	{
		// If the file is read-only, the stream will be read-only
		FILE* file = fopen(pcszStream_, "r+b");
		fReadOnly_ |= !file;
		if (file)
			fclose(file);

#ifdef USE_ZLIB
		// Try and open it as a zip file
		unzFile hfZip;
		if (hfZip = unzOpen(pcszStream_))
		{
			// Iterate through the contents of the zip looking for a file with a suitable size
			for (int nRet = unzGoToFirstFile(hfZip) ; nRet == UNZ_OK ; nRet = unzGoToNextFile(hfZip))
			{
				unz_file_info sInfo;
				char szFile[_MAX_PATH];

				// Get details of the current file
				unzGetCurrentFileInfo(hfZip, &sInfo, szFile, sizeof szFile, NULL, 0, NULL, 0);

				// Continue looking if it's too small to be considered (strictly this shouldn't really be done here!)
				if (sInfo.uncompressed_size < 100000)
					continue;

				// Ok, so open and use the first file in the zip and use that
				if (unzOpenCurrentFile(hfZip) == UNZ_OK)
					return new CZipStream(hfZip, pcszStream_, true/*fReadOnly_*/);	// ZIPs are currently read-only
			}

			// Failed to open the first file, so close the zip
			unzClose(hfZip);
		}
		else
		{
			// Try gzipped or regular file
			gzFile hfGZip;
			if (hfGZip = gzopen(pcszStream_, "rb"))
			{
				// Naughty direct access to a member of gz_stream to determine if the file is compressed
				bool fCompressed = !*reinterpret_cast<int*>(reinterpret_cast<BYTE*>(hfGZip)+88);

				// Only use a CZLibStream object for real compressed data
				if (fCompressed)
					return new CZLibStream(hfGZip, pcszStream_, fReadOnly_);

				gzclose(hfGZip);
#endif	// USE_ZLIB

				// Re-open the file using the regular file functions
				FILE* hf;
				if (hf = fopen(pcszStream_, "rb"))
					return new CFileStream(hf, pcszStream_, fReadOnly_);
#ifdef USE_ZLIB
			}
		}
#endif
	}

	// Couldn't handle what we were given :-/
	return NULL;
}

////////////////////////////////////////////////////////////////////////////////

void CFileStream::Close ()
{
	if (m_hFile)
	{
		fclose(m_hFile);
		m_hFile = NULL;
		m_nMode = modeClosed;
	}
}

bool CFileStream::Rewind ()
{
	rewind(m_hFile);
	return true;
}

LONG CFileStream::Read (void* pvBuffer_, LONG lLen_)
{
	if (m_nMode != modeReading)
	{
		// Close the file, if open for writing
		Close();

		// Open the file for writing, using compression if the source file did
		if (m_hFile = fopen(m_pcszStream, "rb"))
			m_nMode = modeReading;
	}

	LONG lRead = m_hFile ? fread(pvBuffer_, 1, lLen_, m_hFile) : 0;
	return (lRead == -1) ? 0 : lRead;
}

LONG CFileStream::Write (void* pvBuffer_, LONG lLen_)
{
	if (m_nMode != modeWriting)
	{
		// Close the file, if open for reading
		Close();

		// Open the file for writing, using compression if the source file did
		if (m_hFile = fopen(m_pcszStream, "wb"))
			m_nMode = modeWriting;
	}

	return m_hFile ? fwrite(pvBuffer_, 1, lLen_, m_hFile) : 0;
}

////////////////////////////////////////////////////////////////////////////////

#ifdef USE_ZLIB

void CZLibStream::Close ()
{
	if (m_hFile)
	{
		gzclose(m_hFile);
		m_hFile = NULL;
		m_nMode = modeClosed;
	}
}

bool CZLibStream::Rewind ()
{
	return !gzrewind(m_hFile);
}

LONG CZLibStream::Read (void* pvBuffer_, LONG lLen_)
{
	if (m_nMode != modeReading)
	{
		// Close the file, if open for writing
		Close();

		// Open the file for writing, using compression if the source file did
		if (m_hFile = gzopen(m_pcszStream, "rb"))
			m_nMode = modeReading;
	}

	LONG lRead = m_hFile ? gzread(m_hFile, pvBuffer_, lLen_) : 0;
	return (lRead == -1) ? 0 : lRead;
}

LONG CZLibStream::Write (void* pvBuffer_, LONG lLen_)
{
	if (m_nMode != modeWriting)
	{
		// Close the file, if open for reading
		Close();

		// Open the file for writing, using compression if the source file did
		if (m_hFile = gzopen(m_pcszStream, "wb9"))
			m_nMode = modeWriting;
	}

	return m_hFile ? gzwrite(m_hFile, pvBuffer_, lLen_) : 0;
}

////////////////////////////////////////////////////////////////////////////////

void CZipStream::Close ()
{
	if (m_hFile)
	{
		unzCloseCurrentFile(m_hFile);
		unzClose(m_hFile);
		m_hFile = NULL;
	}
}

bool CZipStream::Rewind ()
{
	// There's no zip rewind so we close and re-open the current file
	unzCloseCurrentFile(m_hFile);
	return unzOpenCurrentFile(m_hFile) == UNZ_OK;
}

LONG CZipStream::Read (void* pvBuffer_, LONG lLen_)
{
	return unzReadCurrentFile(m_hFile, pvBuffer_, lLen_);
}

LONG CZipStream::Write (void* pvBuffer_, LONG lLen_)
{
	// Currently there's no support for zip writing
	return 0;
}

#endif	// USE_ZLIB
