#include "stdafx.h"

#include "Util.h"
#include "Options.h"

//	NOTES:
//		Both the SAMBUS and DALLAS clock hardware implementations were done by examing how various
//		programs (MasterDOS, BDOS, QDOS) use them, as I don't own either!
//
//		The SAMDOS clock seems to uses 4 bits for each digit, and invalid values can be written
//		and read back (as used for the QDOS clock check).  It appears that only the digits affected
//		by a tick are updated, so any invalid values written will persist until the digits require
//		updating.  Registers 0xd to 0xf appear to be control registers, but not very much is known
//		about them.  Only the bits used by MasterDOS have been implemented so far.
//
//		More is known about the DALLAS clock as the DS 128(8)7 chips are at http://www.dalsemi.com/
//		It's more complicated that I first thought, but it doesn't appear that the SAM clock circuit
//		gives access to all functions anyway.  The normal digit registers and the user RAM area are
//		available, and BDOS relies on read/write checks on the last byte in user RAM to check the
//		presence of a DALLAS clock.  BDOS also correctly disables and re-enables the clock update
//		(supported) around the point where it's read, and uses a value of 0x80 for register B which
//		also ensures the clock uses BCD mode.
//
//		Both clocks appear to work fine with the software I've used.  The handling of invalid values
//		during a clock update is a bit of an unknown.  The current implementation acts on the pairs
//		for each update treating out of range values to be the largest value that the  pair can store,
//		i.e.  77 seconds is treated as 59 seconds and will tick over to 00, etc.
//
//		The first implementation attempted used the 'struct tm' to store times, which is fine for
//		valid dates but causes problems for invalid ones.  Since both clocks read and modify one
//		digit at a time, the digit modification order becomes important.  Changing 27 to 31
//		would go via 37 if the tens digit was changed first, and 31 to 27 also goes via 37 if the
//		ones digit was changed first.  mktime does handle invalid values, but does smart wrapping
//		which of course isn't reversible.  This lead to doing everything manually, but does mean
//		a correct the day of the week value is missing from the SAMBUS implementation for now...
//
//		Note: The clock values are only updated before they're read or modified, and are advanced
//		by the amount of REAL time that has passed since they were last accessed.  This does mean
//		they are less coupled to the emulation speed.  Perhaps they should be linked to the SAM time
//		that has passed instead, though that would mean the clock would lose sync with the REAL time.
//
////////////////////////////////////////////////////////////////////////////////

// ToDo:	Base class from which separate clock classes are derived?
//			Option to fix clock relative to SAM time, so it doesn't run when SimCoupe isn't?


namespace Clock
{

typedef struct tagSAMTIME
{
	int nSecond10,	nSecond1;
	int nMinute10,	nMinute1;
	int nHour10,	nHour1;

	int	nDay10,		nDay1;
	int nMonth10,	nMonth1;
	int nYear10,	nYear1;
}
SAMTIME;


// SAMBUS data
static SAMTIME stSamBus;			// Last SAMBUS time value
static time_t tSamBusLast;			// Last time we updated the SAM clock
static BYTE abSambusRegs[16];		// The 16 SamBus registers

// DALLAS data
static SAMTIME stDallas;			// Last SAMBUS time value
static time_t tDallasLast;			// Last time we updated the SAM clock
static BYTE abDallasRegs[128];		// DALLAS register and static RAM values (the useful ones)
static BYTE bDallasReg;				// Currently selected DALLAS register


static time_t InitTime (SAMTIME* pst_);
static BYTE SamBusDateVal (BYTE bH_, BYTE bL_);
static void Update (SAMTIME* pst_, time_t* ptLapst_);

////////////////////////////////////////////////////////////////////////////////

bool Init ()
{
	// Set the time on the 2 clocks to the current time
	tSamBusLast = InitTime(&stSamBus);
	tDallasLast = InitTime(&stDallas);

	return true;
}

void Exit ()
{
}


BYTE In (WORD wPort_)
{
	// Strip off the bottom 8 bits (239 for CLOCK_PORT)
	wPort_ >>= 8;

	// SAMBUS clock support?
	if (GetOption(sambus_clock) && wPort_ < 0xfe)
	{
		// Get the update the SAMBUS time, unless the clock update is disabled
		if (!(abSambusRegs[0x0d] & 0x02))
			Update(&stSamBus, &tSamBusLast);
		else
			time(&tSamBusLast);

		switch (wPort_ &= 0xf0)
		{
			case 0x00:	return stSamBus.nSecond1;	// Seconds (ones)
			case 0x10:	return stSamBus.nSecond10;	// Seconds (tens)
			case 0x20:	return stSamBus.nMinute1;	// Minutes (ones)
			case 0x30:	return stSamBus.nMinute10;	// Minutes (tens)
			case 0x40:	return stSamBus.nHour1;		// Hours (ones)
			case 0x50:	return stSamBus.nHour10;	// Hours (tens)
			case 0x60:	return stSamBus.nDay1;		// Days (ones)
			case 0x70:	return stSamBus.nDay10;		// Days (tens)
			case 0x80:	return stSamBus.nMonth1;	// Months (ones)
			case 0x90:	return stSamBus.nMonth10;	// Months (tens)
			case 0xa0:	return stSamBus.nYear1;		// Year (ones)
			case 0xb0:	return stSamBus.nYear10;	// Year (tens)

			// These appear to be control registers
			case 0xd0:							// bit 1 NZ for busy, bit 0 NZ for hold clock
			case 0xe0:
			case 0xf0:							// bit 3 NZ for test mode, bit 2 NZ for 24hr, other bits unknown
				return abSambusRegs[wPort_ >> 4];
		}
	}

	// DALLAS data read
	else if (GetOption(dallas_clock) && wPort_ == 0xff)
	{
		// Get the current SAM time, unless the clock update has been disabled
		if (!(abDallasRegs[0x0b] & 0x80))
			Update(&stDallas, &tDallasLast);
		else
			time(&tDallasLast);

		switch (bDallasReg)
		{
			// Date, time and alarm values are stored in the first 10
			case 0x00:		return SamBusDateVal(stDallas.nSecond10, stDallas.nSecond1);// Seconds
			case 0x01:		return 0xff;												// Seconds alarm (returns 'not set')
			case 0x02:		return SamBusDateVal(stDallas.nMinute10, stDallas.nMinute1);// Minutes
			case 0x03:		return 0xff;												// Minutes alarm (returns 'not set')
			case 0x04:		return SamBusDateVal(stDallas.nHour10, stDallas.nHour1);	// Hours
			case 0x05:		return 0xff;												// Hours alarm (returns 'not set')
			case 0x06:		{ time_t t = time(NULL); return localtime(&t)->tm_wday; }	// Day of the week (unsupported)
			case 0x07:		return SamBusDateVal(stDallas.nDay10, stDallas.nDay1);		// Day of the month
			case 0x08:		return SamBusDateVal(stDallas.nMonth10, stDallas.nMonth1);	// Month
			case 0x09:		return SamBusDateVal(stDallas.nYear10, stDallas.nYear1);	// Year

			// The 4 flag registers
			case 0x0a:		return 0x00;	// Register A: bit 7 zero to signify no update due
			case 0x0b:		return abDallasRegs[bDallasReg];
			case 0x0c:		return 0x00;	// Register C: nothing used yet
			case 0x0d:		return 0x80;	// Register D: bit 7 set indicating battery is good

			default:
			{
				// 0x0e to 0x3f is user RAM
				if (bDallasReg >= 0x0e && bDallasReg <= 0x3f)
					return abDallasRegs[bDallasReg];

				// 0x40 to 0x46 hold the model number and 6 serial number bytes - make one up until known!
				else if (bDallasReg >= 0x40 && bDallasReg <= 0x46)
					return bDallasReg - 0x40;

				// Anything else is (including user RAM taken from the register bank
				else
				{
					TRACE("clock_in: unsupported DALLAS register read (register=%#02x)\n", bDallasReg);
					return abDallasRegs[bDallasReg];
				}
				break;
			}
		}
	}

	// Not present
	return 0xff;
}

void Out (WORD wPort_, BYTE bVal_)
{
	// Strip off the bottom 8 bits (always 239 for CLOCK_PORT)
	wPort_ >>= 8;

	if (GetOption(sambus_clock) && wPort_ < 0xfe)
	{	
		// Get the current SAM time, unless the clock update is on hold
		if (!(abSambusRegs[0x0d] & 0x02))
			Update(&stSamBus, &tSamBusLast);
		else
			time(&tSamBusLast);

		// SAMBUS clock only appears to use the lower 4 bits of the value
		bVal_ &= 0x0f;

		// Make the modification depending on which port was written to
		switch (wPort_ &= 0xf0)
		{
			// SAMBUS clock ports
			case 0x00:	stSamBus.nSecond1 = bVal_;	break;		// Seconds (ones)
			case 0x10:	stSamBus.nSecond10 = bVal_;	break;		// Seconds (tens)
			case 0x20:	stSamBus.nMinute1 = bVal_;	break;		// Minutes (ones)
			case 0x30:	stSamBus.nMinute10 = bVal_;	break;		// Minutes (tens)
			case 0x40:	stSamBus.nHour1 = bVal_;	break;		// Hours (ones)
			case 0x50:	stSamBus.nHour10 = bVal_;	break;		// Hours (tens)
			case 0x60:	stSamBus.nDay1 = bVal_;		break;		// Days (ones)
			case 0x70:	stSamBus.nDay10 = bVal_;	break;		// Days (tens)
			case 0x80:	stSamBus.nMonth1 = bVal_;	break;		// Months (ones)
			case 0x90:	stSamBus.nMonth10 = bVal_;	break;		// Months (tens)
			case 0xa0:	stSamBus.nYear1 = bVal_;	break;		// Year (ones)
			case 0xb0:	stSamBus.nYear10 = bVal_;	break;		// Year (tens)

			// These appear to be control registers
			case 0xd0:	bVal_ &= ~0x02;			// bit 1 NZ for busy
			case 0xe0:
			case 0xf0:		// bit 3 NZ for test mode, bit 2 NZ for 24hr, other bits unknown
				abSambusRegs[wPort_ >> 4] = bVal_;
				break;

			default:
				TRACE("Clock Out(): unhandled SAMBUS clock write (port=%#02x, val=%#02x)\n", wPort_ & 0xf0, bVal_);
		}
	}

	else if (GetOption(dallas_clock) && wPort_ >= 0xfe)
	{
		// Get the current SAM time, unless the clock update has been disabled
		if (!(abDallasRegs[0x0b] & 0x80))
			Update(&stDallas, &tDallasLast);
		else
			time(&tDallasLast);

		switch (wPort_)
		{
			// Select a register
			case 0xfe:
				bDallasReg = bVal_ & 0x7f;
				break;

			// Write a register value
			case 0xff:
				abDallasRegs[bDallasReg] = bVal_;
				TRACE("DALLAS reg %#02x set to %#02x\n", bDallasReg, bVal_);
				break;

			default:
				TRACE("Clock Out(): unhandled DALLAS clock write (port=%#02x, val=%#02x)\n", wPort_, bVal_);
		}
	}
}


// Initialise a SAMTIME structure with the current date/time
static time_t InitTime (SAMTIME* pst_)
{
	// Get current local time and increment day of month so it's one-based
	time_t tNow = time(NULL);
	tm* ptm = localtime(&tNow);

	pst_->nYear10 = ptm->tm_year / 10;
	pst_->nYear1  = ptm->tm_year % 10;

	pst_->nMonth10 = ++ptm->tm_mon / 10;	// Change month from zero-based to one-based
	pst_->nMonth1  = ptm->tm_mon % 10;

	pst_->nDay10 = ptm->tm_mday / 10;
	pst_->nDay1  = ptm->tm_mday % 10;

	pst_->nHour10 = ptm->tm_hour / 10;
	pst_->nHour1  = ptm->tm_hour % 10;

	pst_->nMinute10 = ptm->tm_min / 10;
	pst_->nMinute1  = ptm->tm_min % 10;

	pst_->nSecond10 = ptm->tm_sec / 10;
	pst_->nSecond1  = ptm->tm_sec % 10;

	// Return the time that the structure has been filled with
	return tNow;
}

static void Update (SAMTIME* pst_, time_t* ptLapst_)
{
	// Get the time in more normal components
	int nSecond = pst_->nSecond10 * 10 + pst_->nSecond1;
	int nMinute = pst_->nMinute10 * 10 + pst_->nMinute1;
	int nHour	= pst_->nHour10 * 10 + pst_->nHour1;
	int nDay	= pst_->nDay10 * 10 + pst_->nDay1;
	int nMonth	= pst_->nMonth10 * 10 + pst_->nMonth1;
	int nYear	= pst_->nYear10 * 10 + pst_->nYear1;

	// Work out how many seconds have passed since the last SAM time update
	time_t tNow = time(NULL), tDiff = tNow - *ptLapst_;
	*ptLapst_ = tNow;

	// Add on the change in number of seconds, reducing invalid values to 59 before-hand
	tDiff = (nSecond = min(nSecond, 59) + tDiff) / 60;
	pst_->nSecond1 = (nSecond %= 60) % 10;
	pst_->nSecond10 = nSecond / 10;

	// If there's any time left, consider updating the minutes
	if (tDiff)
	{
		// Add on the change in number of minutes, reducing invalid values to 59 before-hand
		tDiff = (nMinute = min(nMinute, 59) + tDiff) / 60;
		pst_->nMinute1 = (nMinute %= 60) % 10;
		pst_->nMinute10 = nMinute / 10;

		// If there's any time left, consider updating the hours
		if (tDiff)
		{
			// Add on the change in number of hours, reducing invalid values to 23 before-hand
			tDiff = (nHour = min(nHour, 23) + tDiff) / 24;
			pst_->nHour1 = (nHour %= 24) % 10;
			pst_->nHour10 = nHour / 10;

			// Any remaining time is in days and affects the date
			if (tDiff)
			{
				// Limit the month so we know how many days are in the current month
				nMonth = min(nMonth ? nMonth : 1, 12);

				while (1)
				{
					// Table for the number of days in each month, corrected for leap years
					static int abDays[] = { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
					abDays[2] = (nYear & 3) ? 28 : 29;

					// Limit the day in the month so we know how 
					nDay = min(nDay ? nDay : 1, abDays[nMonth]);

					int nAdd = min(abDays[nMonth] - nDay, tDiff);

					// If there's not enough to complete the current month, add it on and finish
					if (nDay + tDiff <= abDays[nMonth])
					{
						pst_->nDay1 = (nDay += tDiff) % 10;
						pst_->nDay10 = nDay / 10;
						break;
					}

					// Complete the current month and set the day back to the first of the month
					tDiff -= abDays[nMonth] - nDay + 1;
					nDay = 1;

					// If we've completed a year, move back to Jan and increment the year
					if (++nMonth > 12)
					{
						nMonth = 1;

						pst_->nYear1 = ++nYear % 10;
						pst_->nYear10 = nYear / 10;
					}

					// Update the month digits as they've changed
					pst_->nMonth1 = nMonth % 10;
					pst_->nMonth10 = nMonth / 10;
				}
			}
		}
	}
}

// The format SAMBUS dates are returned in (BCD or binary) depends on bit 2 of register B
static BYTE SamBusDateVal (BYTE bH_, BYTE bL_)
{
	// Bit is set for binary, reset for BCD
	return (abDallasRegs[0x0b] & 0x02) ? (bH_ * 10) + bL_ : (bH_ << 4) | bL_;
}

};	// namespace Clock
