//
//	Notes:
//		This module needs to stay lean to keep the core CPU emulation fast, but it
//		has to deal with a variety of interrupts, counters and memory contention, so
//		it's starting to get bulkier than I'd have liked.  Still, that's the price
//		you have to pay for more accuracy!
//
//		The order of some of the things in this module is pretty critical, particularly
//		with interrupt timing - Defender is a way to test some of it, and I'll have
//		to dig out the test program I used to test the rest (if I can find it).
//
//	ToDo:
//		Move some private variables into the namespace
//		Generate parity table (and other flags tables?) in Init()
//		Move macros into a header or something?

#include "stdafx.h"
#include "Resource.h"

#include "CPU.h"
#include "Display.h"
#include "Memory.h"
#include "Options.h"
#include "Util.h"
#include "Debug.h"
#include "SAASound.h"
#include "Sound.h"
#include "Input.h"
#include "Frame.h"
#include "IO.h"
#include "FDC.h"
#include "Profile.h"
#include "UI.h"

extern bool fDrawFrame;

#define parity(a) (abParity[a])

// Partity for all 256 8-bit values, pre-shifted into the correct position for the parity flag in F
// Might be worth building this at run-time, along with other arithmetic tables...
const BYTE abParity[256] =
{
	4, 0, 0, 4,  0, 4, 4, 0,  0, 4, 4, 0,  4, 0, 0, 4,
	0, 4, 4, 0,  4, 0, 0, 4,  4, 0, 0, 4,  0, 4, 4, 0,
	0, 4, 4, 0,  4, 0, 0, 4,  4, 0, 0, 4,  0, 4, 4, 0,
	4, 0, 0, 4,  0, 4, 4, 0,  0, 4, 4, 0,  4, 0, 0, 4,
	0, 4, 4, 0,  4, 0, 0, 4,  4, 0, 0, 4,  0, 4, 4, 0,
	4, 0, 0, 4,  0, 4, 4, 0,  0, 4, 4, 0,  4, 0, 0, 4,
	4, 0, 0, 4,  0, 4, 4, 0,  0, 4, 4, 0,  4, 0, 0, 4,
	0, 4, 4, 0,  4, 0, 0, 4,  4, 0, 0, 4,  0, 4, 4, 0,
	0, 4, 4, 0,  4, 0, 0, 4,  4, 0, 0, 4,  0, 4, 4, 0,
	4, 0, 0, 4,  0, 4, 4, 0,  0, 4, 4, 0,  4, 0, 0, 4,
	4, 0, 0, 4,  0, 4, 4, 0,  0, 4, 4, 0,  4, 0, 0, 4,
	0, 4, 4, 0,  4, 0, 0, 4,  4, 0, 0, 4,  0, 4, 4, 0,
	4, 0, 0, 4,  0, 4, 4, 0,  0, 4, 4, 0,  4, 0, 0, 4,
	0, 4, 4, 0,  4, 0, 0, 4,  4, 0, 0, 4,  0, 4, 4, 0,
	0, 4, 4, 0,  4, 0, 0, 4,  4, 0, 0, 4,  0, 4, 4, 0,
	4, 0, 0, 4,  0, 4, 4, 0,  0, 4, 4, 0,  4, 0, 0, 4
};


#define a		regs.AF.B.h_
#define f		regs.AF.B.l_
#define b		regs.BC.B.h_
#define c		regs.BC.B.l_
#define d		regs.DE.B.h_
#define e		regs.DE.B.l_
#define h		regs.HL.B.h_
#define l		regs.HL.B.l_

#define af		regs.AF.W
#define bc		regs.BC.W
#define de		regs.DE.W
#define hl		regs.HL.W

#define a1		regs.AF_.B.h_
#define f1		regs.AF_.B.l_
#define b1		regs.BC_.B.h_
#define c1		regs.BC_.B.l_
#define d1		regs.DE_.B.h_
#define e1		regs.DE_.B.l_
#define h1		regs.HL_.B.h_
#define l1		regs.HL_.B.l_

#define alt_af	regs.AF_.W
#define alt_bc	regs.BC_.W
#define alt_de	regs.DE_.W
#define alt_hl	regs.HL_.W

#define ix		regs.IX.W
#define iy		regs.IY.W
#define sp		regs.SP.W
#define pc		regs.PC.W

#define	r		regs.R
#define	i		regs.I
#define	iff1	regs.IFF1
#define	iff2	regs.IFF2
#define	im		regs.IM


////////////////////////////////////////////////////////////////////////////////
//	H E L P E R   M A C R O S


// True if we are in the main screen area and the address is in system ram
// NOTE - Logic differs slightly from the setting of fContended as this is designed to be used AFTER
// g_nLineCycle has been updated whereas fContended was used BEFORE tstates were calculated
#define	IS_CONTENDED(a)		(fMaybeContended &&		  											\
							 (g_nLineCycle > (BORDER_PIXELS - VIDEO_DELAY)) &&					\
							 (g_nLineCycle <= (BORDER_PIXELS + SCREEN_PIXELS - VIDEO_DELAY)) &&	\
							 (RPAGE(a) < N_PAGES_MAIN))

// Update g_nLineCycle for one memory access
#define	MEM_ACCESS(a)		do {g_nLineCycle += 4;										\
								if (IS_CONTENDED(a))									\
									g_nLineCycle = ROUND(g_nLineCycle, 8);} while (0)

// Update g_nLineCycle for one port access
#define PORT_ACCESS(a)		do {g_nLineCycle += 4;										\
    							if ((a) >= 0xf8)                                      	\
									g_nLineCycle = ROUND(g_nLineCycle, 8);} while (0)


int g_nLine;						// Scan line being generated (0 is the top of the generated display, not the main screen)
int g_nLineCycle;			        // Cycles so far in the current scanline
int	g_nPrevLineCycle;				// Cycles before current instruction began

int nDebugging;

unsigned long tstates;
int InterruptType;					// Type of z80 interupt to deal with

DWORD g_dwCycleCounter;				// Global cycle counter used for various timings

bool fDelayedEI;					// Flag and counter to carry out a delayed EI
bool g_fFlashPhase;					// Mode 1 and 2 flash phase, toggled every 16 frames

bool fMaybeContended;				// True if the main 256x192 area of the screen is contended
bool fContended;					// True if the current instruction is contended

Z80Regs regs;


WORD* pHlIxIy, *pNewHlIxIy;
unsigned int radjust;
int nIntActiveCounter, nMidiOutActiveCounter;			// Counter used for int Active timing
DWORD dwLastTime, dwFPSTime;

namespace CPU
{

bool Init (bool InitMemory_/*=true*/)
{
	// Sanity check the endian of the registers structure
	hl = 1;
	if (h)
		Message(msgFatal, "EEK!  The Z80Regs structure is the wrong endian for this platform!\n");

	// Initialise the Z80 registers (everything zero except ix and iy, which are 0xffff - strange but true!)
	memset(&regs, 0, sizeof regs);
	iy = iy = 0xffff;
	radjust = 0;

	// No index prefix seen yet
	pHlIxIy = pNewHlIxIy = &hl;

	// CPU in normal operating mode, with no active interrupt
	InterruptType = Z80_none;
	nIntActiveCounter = nMidiOutActiveCounter = 0;


	// Counter used to determine when each line should be drawn
	tstates = 0;
	nDebugging = 0;
	g_nLineCycle = g_nPrevLineCycle = 0;

	return (!InitMemory_ || Memory::Init()) && IO::Init(!InitMemory_);
}

void Exit ()
{
	IO::Exit();
	Memory::Exit();
}


// Work out if we're in a vertical part of the screen that may be affected by contention
void UpdateContention ()
{
	fMaybeContended = ((!BORD_SOFF || (vmpr_mode < MODE_3)) &&
		(g_nLine >= TOP_BORDER_LINES && g_nLine < (TOP_BORDER_LINES+SCREEN_LINES)));
}


// Read a byte and update timing
inline BYTE timed_read_byte (WORD addr)
{
	MEM_ACCESS(addr);
	return (read_byte(addr));
}

// Read a word and update timing
inline WORD timed_read_word (WORD addr)
{
	MEM_ACCESS(addr);
	MEM_ACCESS(addr + 1);
	return (read_word(addr));
}

// Write a byte and update timing
//inline void timed_write_byte (WORD addr, BYTE contents)
#define timed_write_byte(addr,contents)	do 		\
{                                               \
	MEM_ACCESS(addr);                           \
	write_byte(addr, contents);                 \
} while (0)

// Write a word and update timing
//inline void timed_write_word (WORD addr, WORD contents)
#define timed_write_word(addr,contents)	do 		\
{                                               \
	MEM_ACCESS(addr);                           \
	MEM_ACCESS((addr) + 1);						\
	write_word(addr, contents);                 \
} while (0)


inline void AddTiming (int nCycles_)
{
	// Add the instruction time to the global cycle counter
	g_dwCycleCounter += nCycles_;

	// If a LINE/FRAME interrupt is active reduce the active time by the instruction length
	if (nIntActiveCounter)
	{
		nIntActiveCounter -= nCycles_;

		// Is the frame/line interrupt fading and no longer visible on the status port?
		if (nIntActiveCounter <= (INT_ACTIVE_TIME - INT_VISIBLE_TIME))
		{
			status_reg |= (STATUS_INT_FRAME|STATUS_INT_LINE);

			// If the active interrupt time has expired, clear the frame/line interrupt
			if (nIntActiveCounter <= 0)
			{
				status_reg_active |= (STATUS_INT_FRAME|STATUS_INT_LINE);
				nIntActiveCounter = 0;
			}
		}
	}

    // Is a MIDI OUT in progress?
    if (lpen & LPEN_TXFMST)
	{
        // Is the MIDI OUT interrupt active, or has a byte been sent so the interrupt will need to be activated?
        if (nMidiOutActiveCounter)
        {
            // Take off the current instruction time
            nMidiOutActiveCounter -= nCycles_;

            // Clear the 'transmitting' bit in LPEN and reset the counter as we're done
            if (nMidiOutActiveCounter <= 0)
            {
                lpen &= ~LPEN_TXFMST;
                nMidiOutActiveCounter = 0;
            }
            // Is it time to end the interrupt?
            else if (nMidiOutActiveCounter <= (MIDI_OVERLAP_TIME - MIDI_INT_ACTIVE_TIME))
            {
                status_reg |= STATUS_INT_MIDIOUT;
                status_reg_active |= STATUS_INT_MIDIOUT;
            }
            // If the byte has been transmitted, signal the MIDI OUT interrupt
            else if (nMidiOutActiveCounter <= MIDI_OVERLAP_TIME)
            {
                status_reg &= ~STATUS_INT_MIDIOUT;
                status_reg_active &= ~STATUS_INT_MIDIOUT;
            }

        }
        // Else MIDI OUT has just occured
        else
        {
            // Start the transmission timing, which also counts down to the MIDI OUT interrupt
			// MIDI output hardware works in steps of 32 tstates, at an offset of 16
            nMidiOutActiveCounter = MIDI_TRANSMIT_TIME + (g_nLineCycle+16) - ROUND(g_nLineCycle+16,32) - 4;
        }
    }

#ifdef _DEBUG
	// Spot unrounded cycle times - shouldn't be any left
	if (nCycles_ & 3)
		DebugBreak();
#endif
}

// The main z80 emulation loop
void ExecuteFrame ()
{
	while (1)
	{
#ifdef _DEBUG
		if (nDebugging)
		{
			TRACE("Cycle=%d, Line=%d, Int=%d, Contend=%d\n", g_nLineCycle, g_nLine-TOP_BORDER_LINES, nIntActiveCounter, fContended);

			g_pDisplay->Update(g_nLine, g_nLineCycle);
			Frame::Redraw();

			TRACE("A=%02x B=%02x C=%02x D=%02x E=%02x H=%02x L=%02x\n", regs.AF.B.h_, regs.BC.B.h_, regs.BC.B.l_, regs.DE.B.h_, regs.DE.B.l_, regs.HL.B.h_, regs.HL.B.l_);
			nDebugging = run_debug(pc);
		}
#endif

		pHlIxIy = pNewHlIxIy;
		pNewHlIxIy = &hl;

		// Fetch instruction and advance program counter
		g_nPrevLineCycle = g_nLineCycle;
		MEM_ACCESS(pc);
		BYTE bOpcode = (InterruptType == Z80_pause) ? OP_NOP : read_byte(pc++);

		radjust++;

		// Execute the instruction
		switch (bOpcode)
		{
#include "Z80ops.h"
		}

		// Add on the instruction time to the line counter and the global counter
		// Have to do it this way as AddTiming is used from other places
		AddTiming(g_nLineCycle - g_nPrevLineCycle);

		// Have we reached the right-hand border area?
		if (g_nLineCycle > (TSTATES_PER_LINE - (BORDER_BLOCKS << 3) - VIDEO_DELAY))
		{
			// Only consider interrupts if one isn't already active
			if (nIntActiveCounter <= 0)
			{
				// Check for a LINE interrupt on the following line
				if ((line_int < SCREEN_LINES) && (g_nLine == (line_int+TOP_BORDER_LINES-1)))
				{
					// Signal the LINE interrupt, and start the interrupt counter
					status_reg &= ~STATUS_INT_LINE;
					status_reg_active &= ~STATUS_INT_LINE;
					nIntActiveCounter = INT_ACTIVE_TIME + (TSTATES_PER_LINE - (BORDER_BLOCKS << 3) - VIDEO_DELAY) - g_nLineCycle + 1;
				}

				// Check for a FRAME interrupt on the last line
				else if (g_nLine == (HEIGHT_LINES-1))
				{
					// Signal a FRAME interrupt, and start the interrupt counter
					status_reg &= ~STATUS_INT_FRAME;
					status_reg_active &= ~STATUS_INT_FRAME;
					nIntActiveCounter = INT_ACTIVE_TIME + (TSTATES_PER_LINE - (BORDER_BLOCKS << 3) - VIDEO_DELAY) - g_nLineCycle + 1;
				}
			}
		}

		// Are there any active interrupts?
		if (status_reg_active != STATUS_INT_NONE)
		{
			// Only process the interrupt if interrupts are enabled (and not delayed after a DI)
			// ... and not in the middle of an indexed instruction
			if (iff1 && (bOpcode != OP_EI) && (bOpcode != OP_DI) && (pNewHlIxIy == &hl))
			{
				// Advance PC if we're stopped on a HALT, as we've got an interrupt it was halted for
				if (bOpcode == OP_HALT)
					pc++;

				// Disable interrupts
				iff1 = iff2 = 0;

				g_nPrevLineCycle = g_nLineCycle;
				g_nLineCycle += 8;

				// What we do next depends on the interrupt mode set
				switch (im)
				{
					case 0:
						// Presumably the same, unless taking the instruction from the bus takes longer
					case 1:
						// Push PC onto the stack, and execute the interrupt handler
						push(pc);
						pc = IM1_INTERRUPT_HANDLER;
						break;

					case 2:
						// Fetch the IM 2 handler address from an address formed from I and 0xff (from the bus)
						// This is done before pushing PC in case it changes the handler location!
						WORD wAddr = timed_read_word((i << 8) | 0xff);

						// Push PC onto the stack, and execute the IM 2 handler
						push(pc);
						pc = wAddr;
				}

				AddTiming(g_nLineCycle - g_nPrevLineCycle);
			}
		}

		// Have we finished the current scanline?
		if (g_nLineCycle >= TSTATES_PER_LINE)
		{
			// Subtract a line's worth of cycles and move to the next line down
			g_nLineCycle -= TSTATES_PER_LINE;
			g_nLine++;

			// Work out if we're in a vertical part of the screen that may be affected by contention
			UpdateContention();

			// Update the input in the centre of the screen, well away from the frame boundary,
			// otherwise the ROM key bounce prevention works can discard keypresses (took AGES to find!)
			if (g_nLine == HEIGHT_LINES/2)
				Input::Update();

			// Finish for now if we've reached the end of the frame
			if (g_nLine == HEIGHT_LINES)
				break;
		}
	}
}


// The main Z80 emulation loop
void Run ()
{
	// Loop until told to quit
	while (UI::CheckEvents())
	{
		// Check for keyboard/joystick/mouse input [now done half way through a frame]
		// Avoids the BASIC keyboard scanner thinking keys have bounced and throwing
		// them away, as was the case for the first keyboard press on the boot screen
////	Input::Update();

		// Execute a single frame's worth, generating the display as it goes
		Frame::Start();
		ProfileStart(CPU);
		ExecuteFrame();
		ProfileEnd();
		Frame::End();

		// Generate any remaining samples for the frame just finished
		Sound::FrameUpdate();

		// Check to see if it's time to turn off floppy drive lights
		Floppy::FrameUpdate();

		// Blit the completed frame to the screen
		if (fDrawFrame)
			Frame::Redraw();

		// Start back at top of top border
		g_nLine = 0;

		// Toggle paper/ink colours every 16 frames for for mode 1 and 2 flashing
		static int nFrame = 0;
		if (!(nFrame++ % 16))
			g_fFlashPhase = !g_fFlashPhase;

		// Deal with non maskable interrupt types
		switch (InterruptType)
		{
			case Z80_none:		// Normal operation
			case Z80_pause:		// CPU paused
				break;

			// Non-maskable interrupt
			case Z80_nmi:
				// Advance PC if we're stopped on a HALT
				if (read_byte(pc) == OP_HALT)
					pc++;

				// Save the interrupt status in iff2 and disable interrupts
				iff2 = iff1;
				iff1 = 0;

				// Guessed timing, but not really important...
				g_nLineCycle += 16;

				// Call NMI handler at address 0x0066
				sp -= 2;
				write_word(sp,pc);
				pc = NMI_INTERRUPT_HANDLER;

				InterruptType = Z80_none;
				break;

			// Z80 reset
			case Z80_reset:
				Init();
				InterruptType = Z80_none;
				break;
		}
	}

	TRACE("Quitting main emulation loop...\n");
}

};	// namespace CPU
