//	Done: documentation, tighter motor and spin-up flags, write-protection, direction flag, missing disk error
//	added READ_ADDRESS, keep track of real head position for step in without update, track can't go below zero
//	id marks for custom disk formats, initialise regs to s:0, t:0, s:1, d:255

//	To-do: real delayed spin-up (including 'hang' when command sent and no disk present), data timeouts, 
//         t-state based index pulses as current value seems too tuned for MasterDOS

#include "stdafx.h"

#include "CPU.h"
#include "Util.h"

#include "CDrive.h"


inline void CDrive::ModifyStatus (BYTE bSet_, BYTE bReset_)
{
	// Reset then set the specified bits
	(m_sRegs.bStatus &= ~bReset_) |= bSet_;

	// If the motor enable bit is set, update the last used time
	if (bSet_ & MOTOR_ON)
		m_dwLastMotorOn = g_dwCycleCounter;
}


BYTE CDrive::In (WORD wPort_)
{
	BYTE bRet;

	// Extract the side number the port refers to (0 or 1)
	int nSide = ((wPort_) >> 2) & 1;

	// Register to read from is the bottom 3 bits of the port
	switch (wPort_ & 0x03)
	{
		case regStatus:
		{
			// Return value is the status byte
			bRet = m_sRegs.bStatus;

			// Type 1 command mode uses more status bits
			if (m_pDisk && !(m_sRegs.bCommand & 0x80))
			{
				// Set the write protect bit if the disk is read-only
				if (m_pDisk->IsReadOnly())
					ModifyStatus(WRITE_PROTECT, 0);

				// If a disk is present, the motor is on, and we're in type 1 status mode, toggle the index pulse
				// status bit periodically to show the disk is present and spinning (not too frequently!)
				static int nPulseCount = 1;
				if (IsMotorOn() && !(++nPulseCount & 0x7))
					bRet |= INDEX_PULSE;
			}

			break;
		}

		case regTrack:
			// Return the current track register value (may not match the current physical head position)
			bRet = m_sRegs.bTrack;
			TRACE("Disk track: returning %#02x\n", bRet);
			break;

		case regSector:
			// Return the current sector register value
			bRet = m_sRegs.bSector;
//			TRACE("Disk sector: returning %#02x\n", byte);
			break;

		case regData:
		{
			// Only consider returning data if some is available
			if ((m_sRegs.bStatus & DRQ) && m_uBuffer)
			{
				// Get the next byte
				m_sRegs.bData = *m_pbBuffer++;

				// Has all the data been read?
				if (!--m_uBuffer)
				{
					// Reset BUSY and DRQ to show we're done
					ModifyStatus(0, BUSY | DRQ);

					// Some commands require additional handling
					switch (m_sRegs.bCommand)
					{
						case READ_ADDRESS:
						case READ_TRACK:
							break;

						case READ_1SECTOR:
							// Set the data read status to include data CRC errors
							ModifyStatus(m_bDataStatus, 0);
							break;

						case READ_MSECTOR:
							// Set the data read status to include data CRC errors, and only continue if ok
							ModifyStatus(m_bDataStatus, 0);
							if (!m_bDataStatus)
							{
								IDFIELD id;
								BYTE bStatus;

								// Move on to the next sector, if there is one and there are no ID field errors
								if (m_pDisk->FindNext(&id, &bStatus))
								{
									ModifyStatus(bStatus, 0);

									if (!bStatus)
									{
										TRACE("FDC: Multiple-sector read moving to sector %d\n", id.bSector);

										// Read the data and set up the pointer and count for it
										m_bDataStatus = m_pDisk->ReadData(m_pbBuffer = m_abBuffer, &m_uBuffer);

										// Signal that more data is available for reading
										ModifyStatus(MOTOR_ON, BUSY | DRQ);
									}
								}
							}
							break;

						default:
							TRACE("Data requested for unknown command type (%d)!\n", m_sRegs.bCommand);
					}
				}

				// Return the data register value
				bRet = m_sRegs.bData;
			}
			else
				bRet = m_sRegs.bData;
		}
	}

	return bRet;
}


void CDrive::Out (WORD wPort_, BYTE bVal_)
{
	// Extract the side number the port refers to (0 or 1)
	int nSide = ((wPort_) >> 2) & 1;

	// Register to write to is the bottom 3 bits of the port
	switch (wPort_ & 0x03)
	{
		case regCommand:
		{
			// Reset the status as we're starting a new command
			m_sRegs.bStatus = 0;

			// Motor on
			ModifyStatus(MOTOR_ON, 0);

			// Type 1 commands have additional status flags set
			if (m_pDisk && !(m_sRegs.bCommand & 0x80))
			{
				// Flag spin-up complete if the spin-up bit is set in a type 1 command
				if (!(bVal_ & FLAG_SPINUP))
					ModifyStatus(SPIN_UP, 0);
			}

			// The main command is taken from the top 2 
			switch (m_sRegs.bCommand = bVal_ & 0xf0)
			{
				// Type I commands

				// Restore disk head to track 0
				case RESTORE:
					ModifyStatus(TRACK00, SPIN_UP);
					m_sRegs.bTrack = m_nHeadPos = 0;

					TRACE("FDC: RESTORE\n");
					break;

				// Seek a track (send the track number to the DATA reg)
				case SEEK:
					// The direction flag is affected by the direction we need to seek
					m_sRegs.fDir = (m_sRegs.bData > m_sRegs.bTrack);

					// Move to the required track
					m_sRegs.bTrack = m_nHeadPos = m_sRegs.bData;

					// If we're now at track zero, set the TRACK00 bit in the status
					if (!m_nHeadPos)
						ModifyStatus(TRACK00, 0);

					TRACE("FDC: SEEK to track %d\n", m_sRegs.bData);
					break;

				// Step using current dir without updating track register
				case STEP_NUPD:
					// If the direction flag is zero, step in
					if (!m_sRegs.fDir)
						m_nHeadPos++;

					// Only step out if we're not already at zero
					else if (m_nHeadPos)
						m_nHeadPos--;

					// The TRACK00 status bit reflects whether we're at track zero
					if (!m_nHeadPos)
					{
						ModifyStatus(TRACK00, 0);

						// Note: the track register is still zeroed, even tho we're in non-update mode!
						m_sRegs.bTrack = 0;			
					}

					TRACE("FDC: STEP_NUPD (%s to track %d (%d))\n", m_sRegs.fDir ? "out" : "in", m_nHeadPos, m_sRegs.bTrack);
					break;

				// Step drive using current direction flags
				case STEP_UPD:
					// If the direction flag is zero, step in
					if (!m_sRegs.fDir)
						m_sRegs.bTrack = ++m_nHeadPos;

					// Only step out if we're not already at zero
					else if (m_nHeadPos)
						m_sRegs.bTrack = --m_nHeadPos;

					// The TRACK00 status bit reflects whether we're at track zero
					if (!m_nHeadPos)
						ModifyStatus(TRACK00, 0);

					TRACE("FDC: STEP_UPD (%s to track %d)\n", m_sRegs.fDir ? "out" : "in", m_nHeadPos);
					break;

				// Step in without updating track register
				case STEP_IN_NUPD:
					// Step the real head in
					m_nHeadPos++;

					// Remember the direction we last moved
					m_sRegs.fDir = false;
					TRACE("FDC: STEP_IN_NUPD (to track %d (%d)\n", m_nHeadPos, m_sRegs.bTrack);
					break;

				// Step in and increment track register
				case STEP_IN_UPD:
					// Move in a track and remember the direction we last moved
					m_sRegs.bTrack = ++m_nHeadPos;
					m_sRegs.fDir = false;
					TRACE("FDC: STEP_IN_UPD (to track %d)\n", m_nHeadPos);
					break;

				// Step out without updating track register
				case STEP_OUT_NUPD:
					// Only decrement the track register if it's not already zero
					if (m_nHeadPos)
						m_nHeadPos--;

					// If we're now at track zero, set the TRACK00 bit in the status and zero the track register
					if (!m_nHeadPos)
					{
						m_sRegs.bTrack = 0;			// Note: the FDC does clear this in non-update mode!
						ModifyStatus(TRACK00, 0);
					}

					// Remember the direction we last moved
					m_sRegs.fDir = true;
					TRACE("FDC: STEP_OUT_NUPD (to track %d (%d)\n", m_nHeadPos, m_sRegs.bTrack);
					break;

				// Step out and decrement track register
				case STEP_OUT_UPD:
					// Only decrement the track register if it's not already zero
					if (m_nHeadPos)
						m_sRegs.bTrack = --m_nHeadPos;

					// If we're now at track zero, set the TRACK00 bit in the status
					if (!m_nHeadPos)
						ModifyStatus(TRACK00, 0);
					else
						ModifyStatus(0, TRACK00);

					// Move in a track and remember the direction we last moved
					m_sRegs.fDir = true;
					TRACE("FDC: STEP_OUT_UPD (to track %d)\n", m_nHeadPos);
					break;


				// Type II Commands

				// Read one or multiple sectors
				case READ_1SECTOR:
				case READ_MSECTOR:
				{
					TRACE("FDC: READ_xSECTOR (from side %d, track %d, sector %d)\n", nSide, m_sRegs.bTrack, m_sRegs.bSector);
					ModifyStatus(BUSY, TRACK00 | DELETED_DATA);

					// Locate the sector, reset busy and signal record not found if we couldn't find it
					if (!FindSector(nSide, m_sRegs.bTrack, m_sRegs.bSector))
						ModifyStatus(RECORD_NOT_FOUND, BUSY);
					else
					{
						// Read the data, reporting anything but CRC errors now, as we can't check the CRC until we reach it at the end of the data on the disk
						if ((m_bDataStatus = m_pDisk->ReadData(m_pbBuffer = m_abBuffer, &m_uBuffer)) & ~CRC_ERROR)
							ModifyStatus(m_bDataStatus, BUSY);
						else
							ModifyStatus(DRQ, 0);
					}
				}
				break;

				// Write one or multiple sectors
				case WRITE_1SECTOR:
				case WRITE_MSECTOR:
					TRACE("FDC: WRITE_xSECTOR\n");
					ModifyStatus(BUSY, TRACK00 | DELETED_DATA);

					// Locate the sector, reset busy and signal record not found if we couldn't find it
					IDFIELD sID;
					if (!FindSector(nSide, m_sRegs.bTrack, m_sRegs.bSector, &sID))
						ModifyStatus(RECORD_NOT_FOUND, BUSY);
					else if (m_pDisk->IsReadOnly())
						ModifyStatus(WRITE_PROTECT, BUSY);
					else
					{
						// Prepare data pointer to receive data, and the amount we're expecting
						m_pbBuffer = m_abBuffer;
						m_uBuffer = 128 << sID.bSize;

						// Signal that data is now requested for writing
						ModifyStatus(DRQ, 0);
					}
					break;


				// Type III Commands

				// Read address, read track, write track
				case READ_ADDRESS:
				{
					// The VL 1772 docs say BUSY should be set here, but SAM Dice relies on it not being set.
					// Either the docs are wrong or a data timeout is relied on to reset BUSY. However, SAM Dice
					// then goes on to read the data, which should no longer be available from the controller!
					ModifyStatus(/*BUSY |*/0, TRACK00 | DELETED_DATA);

					// Read an ID field into our general buffer
					IDFIELD* pId = reinterpret_cast<IDFIELD*>(m_pbBuffer = m_abBuffer);
					BYTE bStatus = ReadAddress(nSide, m_sRegs.bTrack, pId);

					// If successful set up the number of bytes available to read
					if (!(bStatus & TYPE23_ERROR_MASK))
					{
						m_uBuffer = sizeof(IDFIELD);
						ModifyStatus(bStatus|DRQ, 0);
					}

					// Set the error status, resetting BUSY so the client sees the error
					else
					{
						ModifyStatus(bStatus, BUSY);
						m_uBuffer = 0;
					}

					TRACE("FDC: READ_ADDRESS (Track:%d Side:%d Sector:%d Size:%d) returned %#02x\n", pId->bSide, pId->bTrack, pId->bSector, 128 << pId->bSize, bStatus);
				}
				break;

				case READ_TRACK:
					ModifyStatus(BUSY, TRACK00 | DELETED_DATA);

					// Prepare the raw track layout for reading, fail if not found
					if (ReadTrack(nSide, m_sRegs.bTrack, m_pbBuffer = m_abBuffer, m_uBuffer = sizeof m_abBuffer))
						ModifyStatus(DRQ, 0);
					else
					{
						ModifyStatus(RECORD_NOT_FOUND, BUSY);
						m_uBuffer = 0;
					}

					TRACE("FDC: READ_TRACK\n");
					break;

				case WRITE_TRACK:
					TRACE("FDC: WRITE_TRACK\n");
					ModifyStatus(BUSY | DRQ, TRACK00 | DELETED_DATA);

					// Set buffer pointer and count ready to write
					m_pbBuffer = m_abBuffer;
					m_uBuffer = sizeof m_abBuffer;

					// Until the format implementation is finished we'll pretend the disk is write-protected
					ModifyStatus(WRITE_PROTECT, BUSY | DRQ);
					break;


				// Type IV Commands

				// Force interrupt
				case FORCE_INTERRUPT:
					TRACE("FDC: FORCE_INTERRUPT\n");

					// Leave motor on but reset everything else
					m_sRegs.bStatus &= MOTOR_ON;

					// Return to type 1 mode
					m_sRegs.bCommand = 0;
					break;
			}
		}
		break;

		case regTrack:
			TRACE("FDC: Set TRACK to %d\n", bVal_);
			m_sRegs.bTrack = bVal_;
			break;

		case regSector:
			TRACE("FDC: Set SECTOR to %d\n", bVal_);
			m_sRegs.bSector = bVal_;
			break;

		case regData:
		{
			// Store the data value
			m_sRegs.bData = bVal_;

			// Are we expecting any data?
			if ((m_sRegs.bStatus & DRQ) && m_uBuffer)
			{
				// Store the byte
				*m_pbBuffer++ = bVal_;

				// Got all the data we need?
				if (!--m_uBuffer)
				{
					// Reset BUSY and DRQ to show we're done
					ModifyStatus(0, BUSY | DRQ);

					// Some commands require additional handling
					switch (m_sRegs.bCommand)
					{
						case WRITE_1SECTOR:
						case WRITE_MSECTOR:
						{
							TRACE("FDC: Writing full sector: side %d, track %d, sector %d\n", nSide, m_sRegs.bTrack, m_sRegs.bSector);

							UINT uWritten;
							BYTE bStatus = m_pDisk->WriteData(m_abBuffer, &uWritten);
							ModifyStatus(bStatus, BUSY|DRQ);
						}
						break;

						case WRITE_TRACK:
						{
							// Formatting isn't supported yet!

							// Fail the write if the disk is write-protected
//							if (m_pDisk && m_pDisk->IsReadOnly())
								ModifyStatus(WRITE_PROTECT, BUSY);

						}
						break;

					default:
						TRACE("!!! Unexpected data arrived!\n");
					}
				}
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////

// Locate and read the specific sector from the disk
bool CDrive::FindSector (int nSide_, int nTrack_, int nSector_, IDFIELD* pID_/*=NULL*/)
{
	// Assume 'record not found' with zero bytes read, until we discover otherwise
	bool fFound = false;

	// Only check for sectors if there are some to check
	if (m_pDisk && m_pDisk->FindInit(nSide_, nTrack_))
	{
		IDFIELD id;
		BYTE bStatus;

		// Loop through all the sectors in the track
		while (m_pDisk->FindNext(&id, &bStatus))
		{
			// Check to see if the side, track and sector numbers match, and the CRC is correct
			if (id.bSide == nSide_ && id.bTrack == nTrack_ && id.bSector == nSector_ && !bStatus)
			{
				// Valid sector found
				fFound = true;

				// Copy the ID field details if required
				if (pID_)
					*pID_ = id;

				break;
			}
		}
	}

	// Return whether or not we found it
	return fFound;
}

// Find and return the data for the next ID field seen on the spinning disk
BYTE CDrive::ReadAddress (int nSide_, int nTrack_, IDFIELD* pIdField_)
{
	// Assume we won't find a record until we do
	BYTE bRetStatus = RECORD_NOT_FOUND;

	// Only check for sectors if there are some to check
	if (m_pDisk && m_pDisk->FindInit(nSide_, nTrack_))
	{
		// Fetch and advance the disk spin position
		int nSpinPos = m_pDisk->GetSpinPos(true);

		// Find the sector we're currently over
		for (int i = 1 ; i <= nSpinPos ; i++)
			m_pDisk->FindNext(pIdField_, &bRetStatus);
	}

	// Return the find status
	return bRetStatus;
}


// Helper macro for function below
#define Put(x,n)	for (i = 0 ; i < n ; i++, *pbTrack_++ = x);

// Construct the raw track from the information of each sector on the track to make it look real
UINT CDrive::ReadTrack (int nSide_, int nTrack_, BYTE* pbTrack_, UINT uSize_)
{
	BYTE* pbStart = pbTrack_;

	// Start with a clean slate
	memset(pbTrack_, 0, uSize_);

	// Only check for sectors if there are some to check
	if (m_pDisk && m_pDisk->FindInit(nSide_, nTrack_))
	{
		IDFIELD	id;
		BYTE	bStatus;
		int		i;

		// Gap 1 and track header (min 32)
		Put(0x4e, 60);

		// Loop through all the sectors in the track
		while (m_pDisk->FindNext(&id, &bStatus))
		{
			Put(0x00, 12);			// Gap 2 (exactly 12)
			Put(0xa1, 3);			// Gap 2 (exactly 3)
			Put(0xfe, 1);			// ID address mark

			Put(id.bTrack, 1);		// Track number
			Put(id.bSide, 1);		// Disk side
			Put(id.bSector, 1);		// Sector number
			Put(id.bSize, 1);		// Sector size
			Put(id.bCRC1, 1);		// CRC1
			Put(id.bCRC2, 1);		// CRC1

			Put(0x4e, 24);			// Gap 3 (min 24)
			Put(0x00, 12);			// Gap 3 (min 8)
			Put(0xa1, 3);			// Gap 3 (exactly 3)
			Put(0xfb, 1);			// Data address mark

			// Fill in the sector data
			UINT uSize;
			m_pDisk->ReadData(pbTrack_, &uSize);
			pbTrack_ += uSize;

			Put(0xf7, 1);			// Data address mark>
			Put(0x4e, 27);			// Gap 4 (min 16)
		}
	}

	// Return the amount filled in
	return pbTrack_ - pbStart;
}

// Verify the track position on the disk by looking for a sector with the correct track number and a valid CRC
BYTE CDrive::VerifyTrack (int nSide_, int nTrack_)
{
	// Assume we won't find a matching record until we do
	BYTE bRetStatus = RECORD_NOT_FOUND;

	// Only check for sectors if there are some to check
	if (m_pDisk && m_pDisk->FindInit(nSide_, nTrack_))
	{
		IDFIELD	id;
		BYTE	bStatus;

		// Loop through all the sectors in the track
		while (m_pDisk->FindNext(&id, &bStatus))
		{
			// Does the track number match where we are?
			if (id.bTrack == nTrack_)
			{
				// Combine any ID field CRC errors with the returned status
				bRetStatus |= bStatus;

				// If the CRC is correct, we've got a match so return
				if (!bStatus)
				{
					// Clear the 'record not found' bit as we've found a match
					bRetStatus &= ~RECORD_NOT_FOUND;
					break;
				}
			}
		}
	}

	// Return the verify status
	return bRetStatus;
}

namespace Floppy
{

void FrameUpdate ()
{
	g_Drive1.FrameEnd();
	g_Drive2.FrameEnd();
}

};

