//	DirectX video handling for surfaces, screen modes, palettes etc.
//
//	ToDo:
//		Use bit masks to generate data to write to surface.  Currently knows of the
//		two common 16-bit formats, and the known 8/32 format.

#include "stdafx.h"

#include "Video.h"
#include "UI.h"
#include "Util.h"
#include "Options.h"
#include "Frame.h"
#include "Display.h"

extern void ResizeWindow ();


DWORD aulPalette[N_PALETTE_COLOURS];											// SAM palette values in appropriate format for video surface
WORD awY[N_PALETTE_COLOURS], awU[N_PALETTE_COLOURS], awV[N_PALETTE_COLOURS];	// YUV values pre-shifted for overlay surface

LPDIRECTDRAWSURFACE pddsPrimary, pddsFront, pddsBack;
DWORD g_dwColourKey;

CDisplay* g_pDisplay;

namespace Video
{

// These are the pixel formats we currently support, and they're in order to try for
DDPIXELFORMAT aspfOverlay[] =
{
	{ sizeof(DDPIXELFORMAT), DDPF_RGB, 0, 16, 0xf800, 0x07e0, 0x001f, 0 },			// 5-6-5 RGB
	{ sizeof(DDPIXELFORMAT), DDPF_RGB, 0, 16, 0x7c00, 0x03e0, 0x001f, 0 },			// 5-5-5 RGB
	{ sizeof(DDPIXELFORMAT), DDPF_FOURCC, MAKEFOURCC('U','Y','V','Y'), 0,0,0,0,0 },
	{ sizeof(DDPIXELFORMAT), DDPF_FOURCC, MAKEFOURCC('Y','U','Y','2'), 0,0,0,0,0 },
};


LPDIRECTDRAW pdd;
LPDIRECTDRAWPALETTE pddPal;
LPDIRECTDRAWCLIPPER pddClipper;


// function to initialize DirectDraw in windowed mode
bool Init ()
{
	bool fRet = false;

	static fReEnter = false;
	if (fReEnter)
	{
		TRACE("Re-entering Video::Init()\n");
		return true;
	}
	else
	{
		TRACE("Entering Video::Init()\n");
		fReEnter = true;
	}

	// Make sure any existing setup is terminated before we set up again
	Exit();

	// create the main DirectDraw object
    HRESULT hr = DirectDrawCreate(GetOption(videomethod) ? NULL : (LPGUID)DDCREATE_EMULATIONONLY, &pdd, NULL);
    if (FAILED(hr))
		Message(msgError, "DirectDrawCreate() failed with %#08lx", hr);
	else
	{
		// Get the driver capabilites so we know what we need to set up
		DDCAPS ddcaps;
		ddcaps.dwSize = sizeof ddcaps;
		pdd->GetCaps(&ddcaps, NULL);

		// Use exclusive mode for full-screen, or normal mode for windowed
		hr = pdd->SetCooperativeLevel(g_hwnd, GetOption(fullscreen) ? DDSCL_EXCLUSIVE|DDSCL_FULLSCREEN|DDSCL_ALLOWREBOOT : DDSCL_NORMAL);
		if (FAILED(hr))
			Message(msgError, "SetCooperativeLevel() failed with %#08lx", hr);
		else
		{
			// Work out the screen dimensions needed for full screen mode
			int nXres = 320 * GetOption(sourcesize), nYres = 240 * GetOption(sourcesize), nDepth = GetOption(depth);

			// Full screen overlays modes aren't always right (320x240 on mine), so use 640x480x8 instead to be safe.
			// Note that the overlay is still 16-bit, so the bit depth of the background doesn't matter
			if (GetOption(fullscreen) && (GetOption(videomethod) >= 4))
			{
				nXres = 640;
				nYres = 480;
				nDepth = 8;
			}

			// Full screen mode requires a display mode change
			if (GetOption(fullscreen) && FAILED(hr = pdd->SetDisplayMode(nXres, nYres, nDepth)))
			{
				TRACE("!!! Failed to set %dx%dx%d mode!\n", nXres, nYres, nDepth);

				// Try 640x480 if we're using 320x240
				if ((GetOption(sourcesize) == 1) && SUCCEEDED(pdd->SetDisplayMode(640, 480, nDepth)))
				{
					TRACE("Fallen back to 640x480x%d\n", nDepth);
					SetOption(sourcesize, 2);
				}
				else
				{
					// Some resolutions are not available in all colour depths, so fall back to trying 8-bit
					SetOption(depth, nDepth = 8);
					if (FAILED(hr = pdd->SetDisplayMode(nXres, nYres, nDepth)))
					{
						Exit();
						Message(msgError, "SetDisplayMode() failed with %#08lx\n", hr);
						return false;
					}
				}
			}

			// Set up what we need for the primary surface
			DDSURFACEDESC ddsd;
			ZeroMemory(&ddsd, sizeof ddsd);
			ddsd.dwSize = sizeof ddsd;
			ddsd.dwFlags = DDSD_CAPS;
			ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

			// Create the primary surface
			if (FAILED(hr = pdd->CreateSurface(&ddsd, &pddsPrimary, NULL)))
				Message(msgError, "Primary CreateSurface() failed with %#08lx", hr);

			// If we're running in a window we need a clipper to keep the image within the window
			else if (!GetOption(fullscreen) && FAILED(hr = pdd->CreateClipper(0, &pddClipper, NULL)))
				Message(msgError, "CreateClipper() failed with %#08lx", hr);
			else
			{
				// If we're running windowed we need to create and attach a clipper to the primary surface
				if (!GetOption(fullscreen))
				{
					// create a clipper for the primary surface
					if (FAILED(hr = pddClipper->SetHWnd(0, g_hwnd)))
						Message(msgError, "Clipper SetHWnd() failed with %#08lx", hr);
					else if (FAILED(hr = pddsPrimary->SetClipper(pddClipper)))
						Message(msgError, "SetClipper() failed with %#08lx", hr);
				}

				// The working surface is created exactly the size we need, and no larger
				ddsd.dwWidth = ((VISIBLE_BLOCKS + SCREEN_BLOCKS + VISIBLE_BLOCKS) << 3) * GetOption(sourcesize);
				ddsd.dwHeight = VISIBLE_LINES + SCREEN_LINES + VISIBLE_LINES;

				// ToDo: DX3 doesn't support DDCAPS2_WIDESURFACES so we can't create surfaces wider than
				// the display. Because of this we'll have to be careful with 320x240 in fullscreen mode!

				// Try for an overlay surface if we can as they usually have zero CPU hit for stretching
				if ((GetOption(videomethod) >= 4) && !(~ddcaps.dwCaps & (DDCAPS_OVERLAY | DDCAPS_OVERLAYSTRETCH)))
				{
					ddsd.dwSize = sizeof DDSURFACEDESC;
					ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT | DDSD_PIXELFORMAT;
					ddsd.ddsCaps.dwCaps = DDSCAPS_OVERLAY | DDSCAPS_VIDEOMEMORY;

					// There's no reliable way to get the supported formats so we just have to try them until one works!
					for (int i = 0 ; i < (sizeof aspfOverlay / sizeof aspfOverlay[0]) ; i++)
					{
						// Set up the next format to try
						ddsd.ddpfPixelFormat = aspfOverlay[i];

						// Make sure we can create and lock the surface before accepting it
						if (SUCCEEDED(hr = pdd->CreateSurface(&ddsd, &pddsFront, NULL)) &&
							SUCCEEDED(hr = pddsFront->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL)))
						{
							// All ok, so unlock it and continue
							pddsFront->Unlock(ddsd.lpSurface);

							// Shall we use the overlay source as the back buffer? (potentially dangerous!)
							if (GetOption(videomethod) == 5)
								(pddsBack = pddsFront)->AddRef();

							UpdateOverlay();
							break;
						}

						// Can't use that one, so try another...
						TRACE("Overlay failed with format %d (%#08lx)\n", i, hr);
					}

					// If all overlay modes failed, fall back to a non-overlay version in the same memory type
					if (!pddsFront)
						SetOption(videomethod, GetOption(videomethod)-2);
				}

				// Set things up for a regular back buffer
				ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
				ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;

				// If we're not forced to use a video memory back buffer, ask for system memory
				if (GetOption(videomethod) != 2)
					ddsd.ddsCaps.dwCaps |= DDSCAPS_SYSTEMMEMORY;

				// If we're using an overlay, we need a 16-bit back buffer.  Some cards won't let us create a YUV surface (if
				// that's what's being used for the overlay) so we create it in RGB format, as we deal with it directly anyway
				if (GetOption(videomethod) >= 4)
				{
					// 5-5-5 16-bit RGB format
					DDPIXELFORMAT ddpfRGB16 = { sizeof DDPIXELFORMAT, DDPF_RGB, 0, 16, 0x7c00, 0x03e0, 0x001f, 0 };

					// We want a specific surface format
					ddsd.dwFlags |= DDSD_PIXELFORMAT;

					// Try 5-5-5 format first
					ddsd.ddpfPixelFormat = ddpfRGB16;

					// Create the surface, and if that fails setup things up to use 5-6-5 format instead
					if (!pddsBack && FAILED(hr = pdd->CreateSurface(&ddsd, &pddsBack, NULL)))
					{
						ddpfRGB16.dwRBitMask = 0xf800;
						ddpfRGB16.dwGBitMask = 0x07e0;
						ddpfRGB16.dwBBitMask = 0x001f;
						ddsd.ddpfPixelFormat = ddpfRGB16;
					}
				}

				// If not using the overlay source, create the back buffer for rendering
				if (!pddsBack && FAILED(hr = pdd->CreateSurface(&ddsd, &pddsBack, NULL)))
					Message(msgError, "Back buffer CreateSurface() failed with %#08lx", hr);
				else
				{
					// Failed to use an overlay?
					if (!pddsFront)
					{
						// If we couldn't use an overlay we may still benefit from having an extra video memory surface to copy the
						// back buffer to before blitting to the primary surface
						if (GetOption(videomethod) == 3)
						{
							// Check for hardware stretch blitting as it may give us stretching and filtering for free
							// Note: some video drivers LIE about this capability - this includes the Matrox G200 and some ATI drivers at very least
							// Those lying will use some form of software emulation which requires reading from the video surface (VERY SLOW!)
							if (!(~ddcaps.dwFXCaps & (DDFXCAPS_BLTSTRETCHX | DDFXCAPS_BLTSTRETCHY)))
							{
								ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
								ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_VIDEOMEMORY;

								// Create it, but don't worry if it fails as we can do without it
								if (FAILED(hr = pdd->CreateSurface(&ddsd, &pddsFront, NULL)))
									TRACE("!!! Video memory buffer CreateSurface() failed with %#08lx", hr);
							}

							// No stretch blit capabilities in hardware so fall back to just use a system buffer
							else
								SetOption(videomethod, 1);
						}

						// Worst case - only the back buffer and primary buffer exist, so treat the primary as the front buffer
						if (!pddsFront)
							(pddsFront = pddsPrimary)->AddRef();
					}

					// If we tried for a video memory backbuffer but didn't manage it, update the video option to show that
					pddsBack->GetSurfaceDesc(&ddsd);
					if (GetOption(videomethod) == 2 && !(ddsd.ddsCaps.dwCaps & DDSCAPS_VIDEOMEMORY))
						SetOption(videomethod, 1);


					// Ok, let's look at what we're dealing with, as it determines the format we draw in
					pddsFront->GetSurfaceDesc(&ddsd);
					bool fYUV = (ddsd.ddpfPixelFormat.dwFlags & DDPF_FOURCC) != 0;
					bool fPalette = (ddsd.ddpfPixelFormat.dwFlags & DDPF_PALETTEINDEXED8) != 0;
					UINT uBPP = ddsd.ddpfPixelFormat.dwRGBBitCount;

					// Get the current Windows palette, so we can preserve the first and last 10 entries, which are used for
					// standard Windows UI components like menus and dialogue boxes, otherwise it looks crap!
					PALETTEENTRY pal[256];
					if (fPalette)
					{
						HDC hdc = GetDC(NULL);
						GetSystemPaletteEntries(hdc, 0, 256, pal);
						ReleaseDC(NULL, hdc);
					}

					// To handle any 16-bit pixel format we need to build some look-up tables
					WORD awRedTab[256], awGreenTab[256], awBlueTab[256];
					if (uBPP == 16)
					{
						WORD wRMask = WORD(ddsd.ddpfPixelFormat.dwRBitMask);
						WORD wGMask = WORD(ddsd.ddpfPixelFormat.dwGBitMask);
						WORD wBMask = WORD(ddsd.ddpfPixelFormat.dwBBitMask);

						// Create the lookup table
						for (DWORD dw = 0; dw < 256 ; dw++)
						{
							awRedTab[dw] = dw ? WORD(((DWORD(wRMask) * (dw+1)) >> 8) & wRMask) : 0;
							awGreenTab[dw] = dw ? WORD(((DWORD(wGMask * (dw+1))) >> 8) & wGMask) : 0;
							awBlueTab[dw] = dw ? WORD(((DWORD(wBMask) * (dw+1)) >> 8) & wBMask) : 0;
						}
					}

					// Loop through SAM's full palette of 128 colours
					for (int i = 0; i < N_PALETTE_COLOURS ; i++)
					{
						// Convert from SAM palette position to 3-bit RGB
						ULONG ulBlue  = ((i&0x01) << 1| ((i&0x10) >> 2) | ((i&0x08) >> 3));
						ULONG ulRed   = ((i&0x02)     | ((i&0x20) >> 3) | ((i&0x08) >> 3));
						ULONG ulGreen = ((i&0x04) >> 1| ((i&0x40) >> 4) | ((i&0x08) >> 3));

						// Extend the 3-bit values to 8-bit, preserving black as black
						ulBlue  = ulBlue  ? ((ulBlue+1)  << 5) -1 : 0;
						ulRed   = ulRed   ? ((ulRed+1)   << 5) -1 : 0;
						ulGreen = ulGreen ? ((ulGreen+1) << 5) -1 : 0;

						// Write the palette components, leaving the first 10 entries for menu colours etc.
						PALETTEENTRY pe = { static_cast<BYTE>(ulRed), static_cast<BYTE>(ulGreen), static_cast<BYTE>(ulBlue), PC_NOCOLLAPSE };
						pal[i+10] = pe;

						// Using YUV on an overlay?
						if (fYUV)
						{
//							BYTE bY = BYTE(16.0  + (0.257 * ulRed) + (0.504 * ulGreen) + (0.098 * ulBlue));
//							BYTE bU = BYTE(128.0 - (0.147 * ulRed) - (0.289 * ulGreen) + (0.436 * ulBlue));
							BYTE bV = BYTE(128.0 + (0.36  * ulRed) - (0.29  * ulGreen) - (0.07  * ulBlue));

							BYTE bY = BYTE(/*16.0  +*/ (0.299 * ulRed) + (0.587 * ulGreen) + (0.114*ulBlue));
							BYTE bU = BYTE(128.0 + (-0.147 * ulRed) + (-0.289 * ulGreen) + (0.436 * ulBlue));
//							BYTE bV = BYTE(128.0 + (0.615  * ulRed) + (-0.515  * ulGreen) + (-0.100  * ulBlue));

/*
	Also:

	R = Y + (1.4075 * (V-128))
	G = Y - (0.3455 * (U-128) - (0.7169 * (V-128))
	B = Y + (1.779 * (U-128))

	(assumes all values in the range of 0-255)
*/



							// Pre-shift the YUV data for the two formats we currently support
							if (ddsd.ddpfPixelFormat.dwFourCC == MAKEFOURCC('Y','U','Y','2'))
							{
								awY[i] = bY;
								awU[i] = WORD(bU) << 8;
								awV[i] = WORD(bV) << 8;
							}
							else if (ddsd.ddpfPixelFormat.dwFourCC == MAKEFOURCC('U','Y','V','Y'))
							{
								awY[i] = WORD(bY) << 8;
								awU[i] = bU;
								awV[i] = bV;
							}
							else
							{
								TRACE("Unknown YUV FOURCC: %#08lx\n", ddsd.ddpfPixelFormat.dwFourCC);
								DebugBreak();
							}

							aulPalette[i] = (ulBlue<<16) | (ulGreen<<8) | ulRed;
						}

						// In 8 bit mode use offset palette positions to allow for system colours in the first 10
						else if (fPalette)
							aulPalette[i] = i+10;

						else if (uBPP == 16)
							aulPalette[i] = awRedTab[ulRed] | awGreenTab[ulGreen] | awBlueTab[ulBlue];

						// 24-bit mode uses BGR format
						else if (uBPP == 24)
							aulPalette[i] = (ulBlue<<16) | (ulGreen<<8) | ulRed;

						// In 32-bit mode use the full RGB values
						else
							aulPalette[i] = (ulRed<<16) | (ulGreen<<8) | ulBlue;
					}

					// Is the palette required?
					if (fPalette)
					{
						// Create and activate the palette
						if (FAILED(pdd->CreatePalette(DDPCAPS_8BIT, pal, &pddPal, NULL)))
							Message(msgError, "CreatePalette() failed with %#08lx", hr);
						else if (FAILED(pddsPrimary->SetPalette(pddPal)))
							Message(msgError, "SetPalette() failed with %#08lx", hr);
					}

					// Because the pixel format may have changed, we need to refresh the SAM CLUT pixel values
					for (i = 0 ; i < 16 ; i++)
						clut[i] = aulPalette[clutval[i]];

					fRet = UpdateDepth();
					ResizeWindow();
				}
			}
		}
	}

	fReEnter = false;
	TRACE("Leaving Video::Init()\n");
    return fRet;
}

// Cleanup DirectX by releasing all the interfaces we have
void Exit()
{
	TRACE("Video::Exit()\n");

	if (pddPal) { TRACE("Releasing palette\n"); pddPal->Release(); pddPal = NULL; }
	if (pddClipper) { TRACE("Releasing clipper\n"); pddClipper->Release(); pddClipper = NULL; }

	if (pddsFront) { TRACE("Releasing front buffer\n"); pddsFront->Release(); pddsFront = NULL; }
	if (pddsBack) { TRACE("Releasing back buffer\n"); pddsBack->Release(); pddsBack = NULL; }
	if (pddsPrimary) { TRACE("Releasing primary buffer\n"); pddsPrimary->Release(); pddsPrimary = NULL; }

	if (pdd)
	{
		// Should be done automatically, but let's do it just in case
		pdd->RestoreDisplayMode();
		pdd->SetCooperativeLevel(g_hwnd, DDSCL_NORMAL);

		TRACE("Releasing DD\n"); pdd->Release();
		pdd = NULL;
	}
	TRACE("Leaving Video::Exit()\n");
}

void UpdatePalette ()
{
	if (pddPal)
	{
		TRACE("Updating palette\n");
		pddsPrimary->SetPalette(pddPal);
	}
}


bool UpdateDepth ()
{
	Frame::Clear();

	// If a display already exists, delete the object
	if (g_pDisplay)
	{
		delete g_pDisplay;
		g_pDisplay = NULL;
	}


	HRESULT hr;
	DDSURFACEDESC ddsd;
	ddsd.dwSize = sizeof ddsd;
	if (SUCCEEDED(hr = pddsBack->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL)))
	{
		pddsBack->Unlock(ddsd.lpSurface);

		BYTE* pbFrame = reinterpret_cast<BYTE*>(ddsd.lpSurface);
		LONG lPitch = ddsd.lPitch;

		// Get details on the format of the front buffer, which determines how we draw things
		pddsFront->GetSurfaceDesc(&ddsd);

		if (GetOption(drawmethod) == 1)
		{
			if (GetOption(sourcesize) == 1)
				g_pDisplay = new CDisplayXx1<BYTE,F_PALETTE>(NULL, ddsd.dwWidth);
			else
				g_pDisplay = new CDisplayXx1<BYTE,F_DOUBLEX|F_PALETTE>();
		}

		// YUV overlay format?
		else if (ddsd.ddpfPixelFormat.dwFlags & DDPF_FOURCC)
		{
			// All the supported overlays are 16-bit so we only have to check for horizontal doubling
			if (GetOption(sourcesize) == 1)
				g_pDisplay = new CDisplayXx1<WORD,F_YUV>(pbFrame,lPitch);
			else
				g_pDisplay = new CDisplayXx1<WORD,F_DOUBLEX|F_YUV>(pbFrame,lPitch);
		}

		else if (GetOption(sourcesize) == 1)
		{
			switch (ddsd.ddpfPixelFormat.dwRGBBitCount)
			{
				case 8:		g_pDisplay = new CDisplayXx1<BYTE,0>(pbFrame,lPitch);	break;
				case 16:	g_pDisplay = new CDisplayXx1<WORD,0>(pbFrame,lPitch);	break;
				case 32:	g_pDisplay = new CDisplayXx1<DWORD,0>(pbFrame,lPitch);	break;
			}
		}
		else
		{
			switch (ddsd.ddpfPixelFormat.dwRGBBitCount)
			{
				case 8:		g_pDisplay = new CDisplayXx1<BYTE,F_DOUBLEX>(pbFrame,lPitch);	break;
				case 16:	g_pDisplay = new CDisplayXx1<WORD,F_DOUBLEX>(pbFrame,lPitch);	break;
				case 32:	g_pDisplay = new CDisplayXx1<DWORD,F_DOUBLEX>(pbFrame,lPitch);	break;
			}
		}

		if (!g_pDisplay)
			Message(msgError, "Sorry, %d-bit mode is not supported!", ddsd.ddpfPixelFormat.dwRGBBitCount);

		// Restore the original display mode
		g_pDisplay->SetMode((vmpr_mode >> 5) + 1);
	}

	return (g_pDisplay != NULL);
}


void UpdateOverlay ()
{
	TRACE("Updating overlay...\n");

	if (pddsBack && pddsFront && pddsPrimary)
	{
		DDSURFACEDESC ddsdBack, ddsdFront, ddsdPrimary;
		ddsdBack.dwSize = ddsdFront.dwSize = ddsdPrimary.dwSize = sizeof ddsdBack;
		pddsBack->GetSurfaceDesc(&ddsdBack);
		pddsFront->GetSurfaceDesc(&ddsdFront);
		pddsPrimary->GetSurfaceDesc(&ddsdPrimary);

		// Return if the front buffer isn't an overlay surface
		if (!(ddsdFront.ddsCaps.dwCaps & DDSCAPS_OVERLAY))
			return;


		// Determine the screen coordinated for the top-left of the client area
		POINT ptClient = {0,0};
		ClientToScreen(g_hwnd, &ptClient);

		// Where/how the image appears on the screen depends on whether we're fullscreen or in a window
		RECT rFront, rClip;
		if (GetOption(fullscreen))
		{
			int nFit = ddsdPrimary.dwWidth / ddsdBack.dwWidth;
			int nWidth = nFit * (GetOption(ratio5_4) ? ddsdBack.dwWidth * 5 / 4 : ddsdBack.dwWidth);
			int nHeight = nFit * ddsdBack.dwHeight * GetOption(sourcesize);
			TRACE("Back width=%d, back height=%d, primary width=%lu, primary height=%lu\n", nWidth, nHeight, ddsdPrimary.dwWidth, ddsdPrimary.dwHeight);

			// Work out offsets required to centralise the target image (may not fit yet)
			int nOffsetX = ((int)ddsdPrimary.dwWidth - nWidth) >> 1, nOffsetY = ((int)ddsdPrimary.dwHeight - nHeight) >> 1;
			TRACE("Offset X=%d, Offset Y = %d\n", nOffsetX, nOffsetY);
			TRACE("Back width=%d, back height=%d\n", nWidth, nHeight);
			SetRect(&rFront, nOffsetX, nOffsetY, nOffsetX + nWidth, nOffsetY + nHeight);

			// The clip rect is the full screen
			SetRect(&rClip, 0, 0, ddsdPrimary.dwWidth, ddsdPrimary.dwHeight);
		}
		else
		{
			// The target image is to fill the client area of the window (which is already the correct size)
			GetClientRect(g_hwnd, &rFront);
			OffsetRect(&rFront, ptClient.x, ptClient.y);

			// Determine the clip rect for the window, as it may be hanging off the screen edges
			HDC hdc = GetDC(g_hwnd);
			int nClipBox = GetClipBox(hdc, &rClip);
			ReleaseDC(g_hwnd, hdc);
			OffsetRect(&rClip, ptClient.x, ptClient.y);
			TRACE("Clip box = %d,%d,%d,%d\n", rClip.left, rClip.top, rClip.right, rClip.bottom);
		}


		// Clip the front buffer to the clip region to give the final visible area on the screen
		RECT rClipped;
		IntersectRect(&rClipped, &rClip, &rFront);
		TRACE("Front rect: %d,%d,%d,%d\n", rFront.left, rFront.top, rFront.right, rFront.bottom);
		TRACE("Clipped front buffer = %d,%d,%d,%d\n", rClip.left, rClipped.top, rClipped.right, rClipped.bottom);

		// Modify the source rectangle to the region required for the new visble front buffer region
		RECT rBack;
		int nFrontWidth = rFront.right - rFront.left, nFrontHeight = rFront.bottom - rFront.top;
		rBack.left = MulDiv(ddsdBack.dwWidth, rClipped.left-rFront.left, nFrontWidth);
		rBack.top = MulDiv(ddsdBack.dwHeight, rClipped.top-rFront.top, nFrontHeight);
		rBack.right = ddsdBack.dwWidth - MulDiv(ddsdBack.dwWidth, rFront.right-rClipped.right, nFrontWidth);
		rBack.bottom = ddsdBack.dwHeight - MulDiv(ddsdBack.dwHeight, rFront.bottom-rClipped.bottom, nFrontHeight);
		TRACE("Overlay source reduced from %dx%d to %d,%d,%d,%d\n", ddsdBack.dwWidth, ddsdBack.dwHeight, rBack.left, rBack.top, rBack.right, rBack.bottom);

		// Save the pixel from 0,0 on the display and set to what we're hoping to use for the colour-key
		HDC hdc;
		pddsPrimary->GetDC(&hdc);
		COLORREF rgbPrev = GetPixel(hdc, 0, 0);
		SetPixel(hdc, 0, 0, (ddsdPrimary.ddpfPixelFormat.dwFlags & DDPF_PALETTEINDEXED8) ? RGB(0xff,0x00,0xff) : RGB(0x08,0x08,0x08));
		pddsPrimary->ReleaseDC(hdc);

		// Lock the surface and see what the value is for the current mode
		if (SUCCEEDED(pddsPrimary->Lock(NULL, &ddsdPrimary, DDLOCK_SURFACEMEMORYPTR|DDLOCK_WAIT, NULL)))
		{
			g_dwColourKey = *reinterpret_cast<DWORD*>(ddsdPrimary.lpSurface);
			pddsPrimary->Unlock(NULL);

			// Limit the value to the number of bits used for the current depth
			if (ddsdPrimary.ddpfPixelFormat.dwRGBBitCount < 32)
				g_dwColourKey &= (1 << ddsdPrimary.ddpfPixelFormat.dwRGBBitCount) - 1;

			TRACE("Colour key used: %#08lx\n", g_dwColourKey);
		}

		pddsPrimary->GetDC(&hdc);
		SetPixel(hdc, 0, 0, rgbPrev);
		pddsPrimary->ReleaseDC(hdc);


		DDOVERLAYFX ddofx;
		ddofx.dwSize = sizeof DDOVERLAYFX;
		ddofx.dckDestColorkey.dwColorSpaceLowValue = ddofx.dckDestColorkey.dwColorSpaceHighValue = g_dwColourKey;

		// If there's any part of the window to show, set the overlay positions, otherwise hide it
		HRESULT hr = 0;
		if (IsRectEmpty(&rClipped) || FAILED(hr = pddsFront->UpdateOverlay(&rBack, pddsPrimary, &rClipped, DDOVER_SHOW | DDOVER_KEYDESTOVERRIDE, &ddofx)))
		{
			if (FAILED(hr))
				TRACE("!!! Failed to update overlay (%#08lx)\n", hr);

			pddsFront->UpdateOverlay(NULL, pddsPrimary, NULL, DDOVER_HIDE, NULL);
		}
		else
			TRACE("Overlay updated: %d,%d,%d,%d -> %d,%d,%d,%d\n", rBack.left, rBack.top, rBack.right, rBack.bottom, rClipped.left, rClipped.top, rClipped.right, rClipped.bottom);
	}
}


void WaitVBlank ()
{
	pdd->WaitForVerticalBlank(DDWAITVB_BLOCKEND, 0);
}

}	// namespace Video
